@@ -9,6 +9,7 @@ import scipy as sp
 from sklearn.cross_decomposition import PLSRegression
 from sklearn.linear_model import LinearRegression
 from .utils import convex_relaxation
+from scipy.spatial.distance import cdist
 
 
 def ds_pc_transfer_fit(X_primary, X_secondary, max_ncp):
@@ -299,13 +300,12 @@ def epo_fit(X_primary, X_secondary,epo_ncp=1):
     -----
     `E` comes from `SVD(D)` where `D = X_primary - X_secondary`
     Orthogonalization of matrices as:
-        .. math:: X_{pE} = X_p E + a
-        .. math:: X_{sE} = X_s E + a
+        .. math:: X_{pE} = X_p E 
+        .. math:: X_{sE} = X_s E 
     If used for a bilinear model, retrain model using `X_{pE}` and `y`. Obtain model 
         .. math:: y = X_{pE} B_e + \beta_e
-    Then the final regression model after EPO transfer becomes
-        .. math:: B_s = E B_e
-        .. math:: \beta_s = a B_e + \beta_e
+    No further orthogonalization is needed for future predictions.
+  
         
     References
     ----------
@@ -335,8 +335,8 @@ def epo_fit(X_primary, X_secondary,epo_ncp=1):
     >>> pls2.fit(X_primary_e, Y)
     >>> B_pe = pls2.coef_
     >>> beta_pe = pls2.y_mean_ - (x_mean_e.dot(B_pe)) # Model after EPO y = X_se B_pe + beta_pe; X_se = X_s E +a
-    >>> B_s = E.dot(B_pe)
-    >>> beta_s = a.dot(B_pe) + beta_pe # transferred model y = X_s Bs + beta_s
+    >>> B_s = B_pe
+    >>> beta_s = beta_pe 
     >>> print(X_secondary.dot(B_s) + beta_s)
     >>> print(Y)
     [[ 0.47210387]
@@ -810,4 +810,333 @@ def dipals(xcal, ycal, xs, xt, A, l, center_mode = "target", heuristic=False):
         
         beta = b0 - (x_mu_final@B)
 
-        return B, beta, T, Ts, Tt, W, P, Ps, Pt, E, Es, Et, Ey, C, opt_l, discrepancy
\ No newline at end of file
+        return B, beta, T, Ts, Tt, W, P, Ps, Pt, E, Es, Et, Ey, C, opt_l, discrepancy
+    
+    
+
+def sst(X_primary, X_secondary, ncomp):
+    
+    
+    '''    
+    Spectral Space Transformation (SST).
+    Method based on standard samples
+    
+    Parameters
+    ----------    
+    X_primary : ndarray
+        A 2-D array corresponding to the spectral matrix. Shape (N, K) from primary domain
+        
+    X_secondary : ndarray
+        A 2-D array corresponding to the spectral matrix. Shape (N, K) from secondary domain. X_secondary is expected to be paired with X_primary (standard samples)
+        
+    ncomp : int
+        Number of pc's to perform SVD reduction. 
+        
+    Returns
+    -------    
+    out : tuple
+        (F,a), where `F` is the standardization matrix and `a` is the offset.
+        F : ndarray (K,K)
+        a : ndarray (1,K)
+        
+    Notes
+    -----
+    `F` and `a` used as:
+        .. math:: X_p = X_s F + a
+    If used for a bilinear model of the type 
+        .. math:: y = X_p B_p + \beta_p
+    then the final regression model after DS transfer becomes
+        .. math:: B_s = F B_p
+        .. math:: \beta_s = a B_p + \beta_p
+        
+    References
+    ----------
+    W. Du et al., “Maintaining the predictive abilities of multivariate calibration models by spectral space transformation,” Anal. Chim. Acta, vol. 690, no. 1, pp. 64–70, 2011, doi: 10.1016/j.aca.2011.02.014.
+    
+    Examples
+    --------
+    
+    >>> import numpy as np
+    >>> from sklearn.cross_decomposition import PLSRegression
+    >>> import pycaltransfer.caltransfer as caltransfer
+    >>> F_sim = np.array([[0., -0.2, 1.], [1.,0.6,0.8], [0.4,2.5,-1.3]])
+    >>> a_sim = np.array([[2.,5.,4.]])
+    >>> X_secondary = np.array([[0., 0., 1.], [1.,0.,0.], [2.,2.,2.], [2.,5.,4.]])
+    >>> x_error = np.array([[-0.03774524, -0.00475627,  0.01938877],
+    ...       [-0.02925257,  0.1500586 ,  0.01706783],
+    ...       [-0.11049506, -0.03469373, -0.03136003],
+    ...       [-0.00685062, -0.00367186, -0.07211823]])
+    >>> X_primary = X_secondary.dot(F_sim) + a_sim + x_error
+    >>> x_mean = np.mean(X_primary, axis = 0)
+    >>> x_mean.shape = (1,X_primary.shape[1])
+    >>> Y = np.array([[0.1], [0.9], [6.2], [11.9]])
+    >>> y_mean = np.mean(Y, axis = 0)
+    >>> # plsr model primary domain
+    >>> pls2 = PLSRegression(n_components=2,scale=False)
+    >>> pls2.fit(X_primary, Y)
+    >>> B_p = pls2.coef_
+    >>> beta_p = y_mean - (x_mean.dot(B_p)) # baseline model y = X_primary B_p + beta_p
+    # SST transfer
+    >>> F, a = caltransfer.sst(X_primary, X_secondary, ncomp = 2) # with 3 components it's already a perfect fit (possibly overfit)
+    >>> B_ds = F.dot(B_p)
+    >>> beta_s = a.dot(B_p) + beta_p # transferred model y = X_secondary B_s + beta_s
+    >>> print(pls2.predict(X_primary))
+    >>> print(pls2.predict(X_secondary.dot(F) + a))
+        [[ 0.10057381]
+         [ 1.03398846]
+         [ 5.95594099]
+         [12.00949675]]
+        [[ 0.1351422 ]
+         [ 0.71033383]
+         [ 6.49088307]
+         [11.76364089]]
+
+
+    '''
+
+
+    k = X_primary.shape[1]
+    X_primary_mean = np.mean(X_primary, axis=0)
+    X_secondary_mean = np.mean(X_secondary, axis=0)
+    X_primary_mean.shape = (1,X_primary_mean.shape[0])
+    X_secondary_mean.shape = (1,X_secondary_mean.shape[0])
+
+    X_primary_c = X_primary - X_primary_mean
+    X_secondary_c = X_secondary - X_secondary_mean
+
+    X_comb = np.concatenate((X_primary_c, X_secondary_c), axis = 1)
+
+
+    U, S, Vt = np.linalg.svd(X_comb)
+
+    U = U[:,0:ncomp]
+    S = S[0:ncomp]
+    Vt = Vt[0:ncomp,:]
+    V = Vt.T
+
+    Vp = V[0:k,:]
+    Vs = V[k:,:]
+
+
+    F = np.eye(k)+np.linalg.pinv(Vs.T).dot(Vp.T - Vs.T)
+    a = X_primary_mean - (X_secondary_mean.dot(F))
+    
+    return (F,a)
+
+
+
+def top(X_tuple,top_ncp=1):
+    
+    
+    '''
+    Transfer by orthogonal projection (TOP).
+    Method proposed for standardization samples and transfer to multiple devices.
+    
+    Parameters
+    ----------    
+    X_tuple : tuple of ndarray's
+        Each array is a  2-D array corresponding to the spectral matrix. Shape (N, K) for N transfer or standardization samples
+        
+ 
+    top_ncp : int
+        Number of TOP components to remove. 
+          
+    Returns
+    -------    
+    out : ndarray
+        E, where `E` is the orthogonalization matrix.
+        E : ndarray (K,K)
+        
+        
+    Notes
+    -----
+    `E` comes from `SVD(R)` where `R` contains in each row the mean of one instrument.
+    Orthogonalization of matrices as:
+        .. math:: X_{pE} = X_p E 
+        .. math:: X_{sE} = X_s E 
+    If used for a bilinear model, retrain model using `X_{pE}` and `y`. Obtain model 
+        .. math:: y = X_{pE} B_e + \beta_e
+    No further orthogonalization is needed for future predictions.
+  
+        
+    References
+    ----------
+    
+     A. Andrew and T. Fearn, “Transfer by orthogonal projection: Making near-infrared calibrations robust to between-instrument variation,” Chemom. Intell. Lab. Syst., vol. 72, no. 1, pp. 51–56, 2004, doi: 10.1016/j.chemolab.2004.02.004.
+    
+    Examples
+    --------
+    
+    >>> import numpy as np
+    >>> import pycaltransfer.caltransfer as caltransfer
+    >>> from sklearn.cross_decomposition import PLSRegression
+    >>> X_primary = np.array([[0., 0., 1.], [1.,0.,0.], [2.,2.,2.], [2.,5.,4.]])
+    >>> x_error = np.array([[-0.03774524, -0.00475627,  0.01938877],
+    ...       [-0.02925257,  0.1500586 ,  0.01706783],
+    ...       [-0.11049506, -0.03469373, -0.03136003],
+    ...       [-0.00685062, -0.00367186, -0.07211823]])
+    >>> X_secondary = X_primary + x_error
+    >>> # TOP
+    >>> E = caltransfer.top((X_primary, X_secondary),top_ncp=1)
+    >>> X_primary_e = X_primary.dot(E) 
+    >>> x_mean_e = np.mean(X_primary_e, axis = 0)
+    >>> x_mean_e.shape = (1,X_primary_e.shape[1])
+    >>> Y = np.array([[0.1], [0.9], [6.2], [11.9]])
+    >>> y_mean = np.mean(Y, axis = 0)
+    >>> # PLSR after TOP
+    >>> pls2 = PLSRegression(n_components=2,scale=False)
+    >>> pls2.fit(X_primary_e, Y)
+    >>> B_s = pls2.coef_
+    >>> beta_s = y_mean - (x_mean_e.dot(B_s)) # Model after TOP y = X_s B_s + beta_s
+    >>> print(X_secondary.dot(B_s) + beta_s)
+    >>> print(Y)
+        [[ 0.38884864]
+         [ 3.76654062]
+         [ 3.75167162]
+         [12.49809575]]
+        [[ 0.1]
+         [ 0.9]
+         [ 6.2]
+         [11.9]]
+
+ 
+
+    
+    '''  
+    
+    m = len(X_tuple)
+    k = X_tuple[0].shape[1]
+    
+    assert m > top_ncp, "top_ncp must be less than m. m is {:d}".format(m)
+    
+    R = np.zeros((m,k))
+    
+    for mi in range(m):
+        
+        R[mi,:] = X_tuple[mi].mean(axis=0)
+    
+ 
+    U0,S,V0t = np.linalg.svd(R)
+    S_matrix = np.zeros((top_ncp,top_ncp))
+    S_matrix[0:top_ncp,:][:,0:top_ncp] = np.diag(S[0:top_ncp])
+    V = V0t[0:top_ncp].T
+    U = U0[:,0:top_ncp]
+
+    E = np.identity(n=V.shape[0]) - V.dot(V.T)       
+
+    
+    return E
+
+
+
+def dop(Xsou, Ysou, Xtar, Ytar, rho, dop_ncomp = 3):
+    
+    '''
+    Dynamic orthogonal projection (DOP) with Gaussian kernel
+    Standard-free method based on reference values of a target domain
+    
+    Parameters
+    ----------
+    Xsou: numpy array (N,K)
+        Spectral calibration data from source domain
+
+    Ysou: numpy array (N,1)
+        Reference values for calibration from source domain
+
+    Xtar: numpy array (N_t,K) 
+        Spectral data from target domain
+
+    Ytar: numpy array (N_t,1)
+        Reference values from target domain
+        
+    rho: float
+        Precision parameter for Gaussian kernel
+        
+    dop_ncomp : int
+        Number of DOP components to remove. 
+          
+    Returns
+    -------    
+    out : ndarray
+        E, where `E` is the orthogonalization matrix.
+        E : ndarray (K,K)
+        
+        
+    Notes
+    -----
+    `E` comes from `SVD(D)` with `D = Xtar - Xtar_hat` where `Xtar_hat` is the matrix with supervised virtual standard samples
+    Orthogonalization of matrices as:
+        .. math:: X_{pE} = X_p E 
+        .. math:: X_{sE} = X_s E 
+    If used for a bilinear model, retrain model using `X_{pE}` and `y`. Obtain model 
+        .. math:: y = X_{pE} B_e + \beta_e
+    No further orthogonalization is needed for future predictions.
+  
+        
+    References
+    ----------
+    
+    M. Zeaiter, J. M. Roger, and V. Bellon-Maurel, “Dynamic orthogonal projection. A new method to maintain the on-line robustness of multivariate calibrations. Application to NIR-based monitoring of wine fermentations,” Chemom. Intell. Lab. Syst., vol. 80, no. 2, pp. 227–235, 2006, doi: 10.1016/j.chemolab.2005.06.011.
+
+
+    Examples
+    --------
+    
+    >>> import numpy as np
+    >>> import pycaltransfer.caltransfer as caltransfer
+    >>> from sklearn.cross_decomposition import PLSRegression
+    >>> Xsou = np.array([[0., 0., 1.], [1.,0.,0.], [2.,2.,2.], [2.,5.,4.]])
+    >>> x_error = np.array([[-0.03774524, -0.00475627,  0.01938877],
+    ...       [-0.02925257,  0.1500586 ,  0.01706783],
+    ...       [-0.11049506, -0.03469373, -0.03136003],
+    ...       [-0.00685062, -0.00367186, -0.07211823]])
+    >>> Xtar = Xsou + x_error
+    >>> Ysou = np.array([[0.1], [0.9], [6.2], [11.9]])
+    >>> Ytar = np.array([[0.4], [0.7], [6], [13]])
+    >>> # DOP
+    >>> E = caltransfer.dop(Xsou, Ysou, Xtar, Ytar, rho = 10, dop_ncomp = 1)
+    >>> Xsou_e = Xsou.dot(E) 
+    >>> x_mean_e = np.mean(Xsou_e, axis = 0)
+    >>> x_mean_e.shape = (1,Xsou_e.shape[1])
+    >>> y_mean = np.mean(Ysou, axis = 0)
+    >>> # PLSR after EPO
+    >>> pls2 = PLSRegression(n_components=2,scale=False)
+    >>> pls2.fit(Xsou_e, Ysou)
+    >>> B_s = pls2.coef_
+    >>> beta_s = y_mean - (x_mean_e.dot(B_s)) # Model after TOP y = X_s B_s + beta_s
+    >>> print(Xtar.dot(B_s) + beta_s)
+    >>> print(Ytar)
+        [[ 0.7848724 ]
+         [ 0.46202639]
+         [ 5.87275856]
+         [11.83470517]]
+        [[ 0.4]
+         [ 0.7]
+         [ 6. ]
+         [13. ]]
+
+    
+    '''
+    
+    Ysouq = Ysou.copy()
+    if Ysouq.ndim < 2:
+        Ysouq.shape = (Ysou.shape[0],1)
+    
+    Ytarq = Ytar.copy()
+    if Ytarq.ndim < 2:    
+        Ytarq.shape = (Ytar.shape[0],1)
+    
+
+
+    F = np.exp(-rho*cdist(Ytarq, Ysouq, metric='euclidean'))
+    denom = F.sum(axis=1)
+    denom.shape = (denom.shape[0],1)
+    F /= np.tile(denom,(1,F.shape[1]))
+
+
+    Xtar_hat = F.dot(Xsou)
+
+    E,a = epo_fit(Xtar_hat, Xtar, epo_ncp = dop_ncomp)
+    
+    return E
+
